-- GridConfigurableLayouts.lua
--
-- Created By : msaint


--{{{ Hack the localization function for easy implementation later
local L = {}
setmetatable(L,{ __index = function(l, text) return text end})
---}}}
     
     
--{{{ Constants
--}}}


--{{{ Local Globals :-P
local combatLockdown = false
local isRegistered = false
local rosterZoneList = {}
local currentLayout = nil
local emptyLayout = {
   defaults = {
  	   unitsPerColumn = 5,
	   maxColumns = 8,
	},
	[1] = {
		nameList = "",
	},
}
local groupList = {}
local outerFrames = {}
local outerPetFrames = {}
local allPetsEnabled = true
---}}}


--{{{ Libraries
local talentLib = LibStub:GetLibrary("LibGroupTalents-1.0")
local GridRoster = Grid:GetModule("GridRoster")
local GridLayout = Grid:GetModule("GridLayout")
local LoadLayoutHook = AceLibrary("AceAddon-2.0"):new("AceHook-2.1")
local waterfall = AceLibrary("Waterfall-1.0")
local media = LibStub("LibSharedMedia-3.0", true)
--}}}


--{{{ Object creation
GridConfigurableLayouts = Grid:NewModule("GridConfigurableLayouts")
GridConfigurableLayouts.menuName = L["Grid Configurable Layouts"]
---}}}


---{{{ Module-wide tables and variables
GridConfigurableLayouts.allPetsEnabled = true
GridConfigurableLayouts.groupTypes = {
   ["tank"] = L["Tanks"],
   ["blizztank"] = L["Blizzard Tanks"],
   ["melee"] = L["Melee"],
   ["healer"] = L["Healers"],
   ["ranged"] = L["Ranged"],
   ["ooz"] = L["Out Of Zone"],
   ["rg1"] = L["Raid Group 1"],
   ["rg2"] = L["Raid Group 2"],
   ["rg3"] = L["Raid Group 3"],
   ["rg4"] = L["Raid Group 4"],
   ["rg5"] = L["Raid Group 5"],
   ["rg6"] = L["Raid Group 6"],
   ["rg7"] = L["Raid Group 7"],
   ["rg8"] = L["Raid Group 8"],
   ["pets"] = L["Pets"],
   ["empty"] = L["Empty"],
   ["none"] = L["None"],
   ["guild"] = L["Guild"],   
}
GridConfigurableLayouts.layoutNames = {}
---}}}


---{{{ Local functions

-- This function borrowed directly from GridLayout.lua
local function findVisibleUnitFrame(f)
	if f:IsVisible() and f:GetAttribute("unit") then
		return f
	end

	for i = 1, select('#', f:GetChildren()) do
		local child = select(i, f:GetChildren())
		local good = findVisibleUnitFrame(child)

		if good then
			return good
		end
	end
end
---}}}


--{{{ AceDB defaults
GridConfigurableLayouts.defaultDB = {
   checkOOZFrequency = 5,
   configLayout = "GCL Default",
   createBorders = true,
   filterToColor = "tank",
   filterColors = {
      ["tank"] = {["r"] = 0.9, ["g"] = 0.5, ["b"] = 0.5, ["a"] = 0.8},
      ["blizztank"] = {["r"] = 0.9, ["g"] = 0.5, ["b"] = 0.5, ["a"] = 0.8},
      ["melee"] = {["r"] = 0.9, ["g"] = 0.9, ["b"] = 0.5, ["a"] = 0.8},
      ["healer"] = {["r"] = 0.5, ["g"] = 0.9, ["b"] = 0.5, ["a"] = 0.8},
      ["ranged"] = {["r"] = 0.5, ["g"] = 0.5, ["b"] = 0.9, ["a"] = 0.8},
      ["ooz"] = {["r"] = 0.5, ["g"] = 0.5, ["b"] = 0.5, ["a"] = 0.5},
      ["rg1"] = {["r"] = 0.5, ["g"] = 0.5, ["b"] = 0.5, ["a"] = 0.5},
      ["rg2"] = {["r"] = 0.5, ["g"] = 0.5, ["b"] = 0.5, ["a"] = 0.5},
      ["rg3"] = {["r"] = 0.5, ["g"] = 0.5, ["b"] = 0.5, ["a"] = 0.5},
      ["rg4"] = {["r"] = 0.5, ["g"] = 0.5, ["b"] = 0.5, ["a"] = 0.5},
      ["rg5"] = {["r"] = 0.5, ["g"] = 0.5, ["b"] = 0.5, ["a"] = 0.5},
      ["rg6"] = {["r"] = 0.5, ["g"] = 0.5, ["b"] = 0.5, ["a"] = 0.5},
      ["rg7"] = {["r"] = 0.5, ["g"] = 0.5, ["b"] = 0.5, ["a"] = 0.5},
      ["rg8"] = {["r"] = 0.5, ["g"] = 0.5, ["b"] = 0.5, ["a"] = 0.5},
      ["pets"] = {["r"] = 0.5, ["g"] = 0.5, ["b"] = 0.5, ["a"] = 0.5},
      ["empty"] = {["r"] = 0.5, ["g"] = 0.5, ["b"] = 0.5, ["a"] = 0.5},
      ["none"] = {["r"] = 0.5, ["g"] = 0.5, ["b"] = 0.5, ["a"] = 0.5},
      ["guild"] = {["r"] = 0.3, ["g"] = 0.9, ["b"] = 0.6, ["a"] = 0.5},
   },
   layouts = {
      ["GCL Default"] = {
         allowRepeats = false,
         numGroups = 6,
         groupType = {"tank", "melee", "healer", "ranged", "pets", "ooz", "none", "none", "none", "none", "none" },
         nameList = {"", "", "", "", "", "", "", "", "", "", "" },
         petOptions = {
            ["allPets"] = false,
            ["HUNTER"] = true,
            ["WARLOCK"] = true,
            ["DEATHKNIGHT"] = true,
         },
      },
      ["By Group 10 w/ Tanks, Pets"] = {
         allowRepeats = false,
         numGroups = 5,
         groupType = {"tank", "rg1", "rg2", "pets", "ooz", "none", "none", "none", "none", "none", "none" },
         nameList = {"", "", "", "", "", "", "", "", "", "", "" },
         petOptions = {
            ["allPets"] = true,
            ["HUNTER"] = true,
            ["WARLOCK"] = true,
            ["DEATHKNIGHT"] = true,
         },
      },
      ["By Group 25 w/ Tanks, Pets"] = {
         allowRepeats = false,
         numGroups = 8,
         groupType = {"tank", "rg1", "rg2", "rg3", "rg4", "rg5", "pets", "ooz", "none", "none", "none" },
         nameList = {"", "", "", "", "", "", "", "", "", "", "" },
         petOptions = {
            ["allPets"] = true,
            ["HUNTER"] = true,
            ["WARLOCK"] = true,
            ["DEATHKNIGHT"] = true,
         },
      }, 
   },
}
--}}}


--{{{ Options Structure
local configurableLayoutsOptions = {
   ["layoutSelectionHeader"] = { 
	   type = "header",
	   name = L["Layout Selection"],
	   order = 1,
	},
   ["configLayout"] = {
      type = "text",
  	   name = L["Layout to configure:"],
  	   desc = L["Sets which layout configuration changes apply to."],
  	   order = 2,
  	   get = function () return GridConfigurableLayouts.db.profile.configLayout end,
  	   set = function (v) GridConfigurableLayouts:loadNewConfig(v) end,
  	   validate = GridConfigurableLayouts.layoutNames,
   },
   ["editGroups"] = {
      type = "group",
      name = L["Edit Group Settings"],
      desc = L["Edit filters to apply to groups in the current layout."],
      order = 3,
      args = {
         ["allowRepeats"] = {
        	   type = "toggle",
           	name = L["Allow repeats"],
           	desc = L["Show raid members in all groups that apply, instead of only once"],
           	order = 1,
           	get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].allowRepeats end,
           	set = function (v)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].allowRepeats =
                  not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].allowRepeats
               GridConfigurableLayouts:checkUpdateCurrent()
            end, 
         },
         ["offlineIsOOZ"] = {
        	type = "toggle",
           	name = L["Treat offline as out of zone"],
           	desc = L["Put raid members that are offline in out of zone group, if present."],
           	order = 1,
           	get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].offlineIsOOZ end,
           	set = function (v)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].offlineIsOOZ =
                  not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].offlineIsOOZ
               GridConfigurableLayouts:checkUpdateCurrent()
            end, 
         },
         ["groupType1"] = {
            type = "text",
            name = L["Group 1:"],
        	   desc = L["Sets contents of the first grid group."],
        	   order = 2,
        	   get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[1] end,
        	   set = function (v)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[1] = v
               GridConfigurableLayouts:checkUpdateCurrent()
            end,
        	   validate = GridConfigurableLayouts.groupTypes,
         },
         ["groupType2"] = {
            type = "text",
        	   name = L["Group 2:"],
        	   desc = L["Sets contents of the second grid group."],
        	   order = 3,
        	   get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[2] end,
        	   set = function (v)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[2] = v
               GridConfigurableLayouts:checkUpdateCurrent()
            end,
        	   validate = GridConfigurableLayouts.groupTypes,
         },
         ["groupType3"] = {
            type = "text",
        	   name = L["Group 3:"],
        	   desc = L["Sets contents of the third grid group."],
        	   order = 5,
        	   get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[3] end,
        	   set = function (v)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[3] = v
               GridConfigurableLayouts:checkUpdateCurrent()
            end,
           	validate = GridConfigurableLayouts.groupTypes,
         },
         ["groupType4"] = {
            type = "text",
        	   name = L["Group 4:"],
        	   desc = L["Sets contents of the fourth grid group."],
        	   order = 6,
        	   get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[4] end,
        	   set = function (v)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[4] = v
               GridConfigurableLayouts:checkUpdateCurrent()
            end,
           	validate = GridConfigurableLayouts.groupTypes,
         },
         ["groupType5"] = {
            type = "text",
        	   name = L["Group 5:"],
        	   desc = L["Sets contents of the fifth grid group."],
        	   order = 7,
        	   get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[5] end,
        	   set = function (v)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[5] = v
               GridConfigurableLayouts:checkUpdateCurrent()
            end,
           	validate = GridConfigurableLayouts.groupTypes,
         },
         ["groupType6"] = {
            type = "text",
        	   name = L["Group 6:"],
        	   desc = L["Sets contents of the sixth grid group."],
        	   order = 8,
        	   get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[6] end,
        	   set = function (v)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[6] = v
               GridConfigurableLayouts:checkUpdateCurrent()
            end,
           	validate = GridConfigurableLayouts.groupTypes,
         },
         ["groupType7"] = {
            type = "text",
        	   name = L["Group 7:"],
        	   desc = L["Sets contents of the seventh grid group."],
        	   order = 9,
        	   get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[7] end,
        	   set = function (v)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[7] = v
               GridConfigurableLayouts:checkUpdateCurrent()
            end,
           	validate = GridConfigurableLayouts.groupTypes,
         },
         ["groupType8"] = {
            type = "text",
        	   name = L["Group 8:"],
           	desc = L["Sets contents of the eighth grid group."],
        	   order = 10,
           	get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[8] end,
           	set = function (v)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[8] = v
               GridConfigurableLayouts:checkUpdateCurrent()
            end,
           	validate = GridConfigurableLayouts.groupTypes,
         },
         ["groupType9"] = {
            type = "text",
        	   name = L["Group 9:"],
        	   desc = L["Sets contents of the ninth grid group."],
        	   order = 11,
        	   get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[9] end,
        	   set = function (v)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[9] = v
               GridConfigurableLayouts:checkUpdateCurrent()
            end,
        	   validate = GridConfigurableLayouts.groupTypes,
         },
         ["groupType10"] = {
            type = "text",
        	   name = L["Group 10:"],
        	   desc = L["Sets contents of the tenth grid group."],
        	   order = 12,
        	   get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[10] end,
        	   set = function (v)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[10] = v
               GridConfigurableLayouts:checkUpdateCurrent()
            end,
           	validate = GridConfigurableLayouts.groupTypes,
         },
         ["groupType11"] = {
            type = "text",
        	   name = L["Group 11:"],
        	   desc = L["Sets contents of the eleventh grid group."],
        	   order = 13,
        	   get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[11] end,
        	   set = function (v)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[11] = v
               GridConfigurableLayouts:checkUpdateCurrent()
            end,
        	   validate = GridConfigurableLayouts.groupTypes,
         },
         ["addGroup"] = {
         type = "execute",
         name = L["Add a Group"],
         desc = L["Add a group."],
         order = 14,
         disabled = function () return (GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups > 10) end,
         func = function ()
            num = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups + 1
            GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[num] = "none"
            GridLayout.options.args["GridConfigurableLayouts"].args["editGroups"].args["groupType"..num].type = "text"
            GridLayout.options.args["GridConfigurableLayouts"].args["editNameLists"].args["nameList"..num].type = "text"
            GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups = num
            waterfall:Refresh("Grid")
         end,
         },
         ["removeGroup"] = {
         type = "execute",
         name = L["Remove a Group"],
         desc = L["Remove a group."],
         order = 15,
         disabled = function () return (GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups < 2) end,
         func = function ()
            num = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups
            GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[num] = "none"
            GridLayout.options.args["GridConfigurableLayouts"].args["editGroups"].args["groupType"..num].type = "disabled"
            GridLayout.options.args["GridConfigurableLayouts"].args["editNameLists"].args["nameList"..num].type = "disabled"
            GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups = num - 1
            waterfall:Refresh("Grid")
         end,
         },    
      },
   },
   ["editNameLists"] = {
      type = "group",
      name = L["Edit Name Lists"],
      desc = L["Edit name lists to apply to groups in the current layout."],
      order = 20,
      args = {
         ["mustHitReturnLabel"] = {
            type = "label",
            name = L["Note: When entering a namelist, you must hit return for the namelist to be registered."],
            order = 22,
         },   
         ["nameList1"] = {
            type = "text",
  	         name = L["Group 1:"],
  	         desc = L["Sets name list for the first grid group (comma separated list)."],
  	         order = 23,
  	         get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[1] end,
  	         set = function (v)
               v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[1] = v
               GridConfigurableLayouts:checkUpdateCurrent()
            end,
         },
         ["nameList2"] = {
            type = "text",
    	      name = L["Group 2:"],
    	      desc = L["Sets name list for the second grid group (comma separated list)."],
    	      order = 24,
    	      get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[2] end,
    	      set = function (v)
                v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[2] = v
                GridConfigurableLayouts:checkUpdateCurrent()
            end,
         },
         ["nameList3"] = {
            type = "text",
            name = L["Group 3:"],
            desc = L["Sets name list for the third grid group (comma separated list)."],
            order = 25,
            get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[3] end,
            set = function (v)
               v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[3] = v
               GridConfigurableLayouts:checkUpdateCurrent()
            end,
         },
         ["nameList4"] = {
             type = "text",
             name = L["Group 4:"],
             desc = L["Sets name list for the fourth grid group (comma separated list)."],
             order = 26,
             get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[4] end,
             set = function (v)
                v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[4] = v
                GridConfigurableLayouts:checkUpdateCurrent()
             end,
         },
         ["nameList5"] = {
             type = "text",
             name = L["Group 5:"],
             desc = L["Sets name list for the fifth grid group (comma separated list)."],
             order = 27,
             get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[5] end,
             set = function (v)
                v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[5] = v
                GridConfigurableLayouts:checkUpdateCurrent()
             end,
         },
         ["nameList6"] = {
             type = "text",
             name = L["Group 6:"],
             desc = L["Sets name list for the sixth grid group (comma separated list)."],
             order = 28,
             get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[6] end,
             set = function (v)
                v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[6] = v
                GridConfigurableLayouts:checkUpdateCurrent()
             end,
         },
         ["nameList7"] = {
             type = "text",
             name = L["Group 7:"],
             desc = L["Sets name list for the seventh grid group (comma separated list)."],
             order = 29,
             get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[7] end,
             set = function (v)
                v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[7] = v
                GridConfigurableLayouts:checkUpdateCurrent()
             end,
         },
         ["nameList8"] = {
             type = "text",
             name = L["Group 8:"],
             desc = L["Sets name list for the eighth grid group (comma separated list)."],
             order = 30,
             get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[8] end,
             set = function (v)
                v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[8] = v
                GridConfigurableLayouts:checkUpdateCurrent()
             end,
         },
         ["nameList9"] = {
             type = "text",
             name = L["Group 9:"],
             desc = L["Sets name list for the ninth grid group (comma separated list)."],
             order = 31,
             get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[9] end,
             set = function (v)
                v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[9] = v
                GridConfigurableLayouts:checkUpdateCurrent()
             end,
         },
         ["nameList10"] = {
             type = "text",
             name = L["Group 10:"],
             desc = L["Sets name list for the tenth grid group (comma separated list)."],
             order = 32,
             get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[10] end,
             set = function (v)
                v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[10] = v
                GridConfigurableLayouts:checkUpdateCurrent()
             end,
         },
         ["nameList11"] = {
             type = "text",
             name = L["Group 11:"],
             desc = L["Sets name list for the eleventh grid group (comma separated list)."],
             order = 33,
             get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[11] end,
             set = function (v)
                  v = v:gsub("(%a)(%a+)", function (a,b) return a:upper() .. b:lower() end)
                  GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[11] = v
                  GridConfigurableLayouts:checkUpdateCurrent()
             end,
         },
         ["addGroup"] = {
         type = "execute",
         name = L["Add a Group"],
         desc = L["Add a group."],
         order = 34,
         disabled = function () return (GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups > 10) end,
         func = function ()
            num = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups + 1
            GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[num] = "none"
            GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[num] = ""
            GridLayout.options.args["GridConfigurableLayouts"].args["editGroups"].args["groupType"..num].type = "text"
            GridLayout.options.args["GridConfigurableLayouts"].args["editNameLists"].args["nameList"..num].type = "text"
            waterfall:Refresh("Grid")
         end,
         },
         ["removeGroup"] = {
         type = "execute",
         name = L["Remove a Group"],
         desc = L["Remove a group."],
         order = 35,
         disabled = function () return (GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups < 2) end,
         func = function ()
            num = GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups
            GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].groupType[num] = "none"
            GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].nameList[num] = ""
            GridLayout.options.args["GridConfigurableLayouts"].args["editGroups"].args["groupType"..num].type = "disabled"
            GridLayout.options.args["GridConfigurableLayouts"].args["editNameLists"].args["nameList"..num].type = "disabled"
            GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].numGroups = num - 1
            waterfall:Refresh("Grid")
         end,
         },
      },
   },
   ["spacer1"] = {
		type = "header",
		name = L[" "],
		order = 50,
	},
	["spacer2"] = {
		type = "header",
		name = L[" "],
		order = 51,
	},
	["spacer3"] = {
		type = "header",
		name = L[" "],
		order = 52,
	},
    ["warningHit Enter"] = { 
	   type = "header",
	   name = L["Note: You must press enter after typing in a text box."],
	   order = 60,
	},
	["spacer4"] = {
		type = "header",
		name = L[" "],
		order = 59,
	},
    ["spacer5"] = {
		type = "header",
		name = L[" "],
		order = 59,
	},
    ["spacer6"] = {
		type = "header",
		name = L[" "],
		order = 59,
	},
    ["spacer7"] = {
		type = "header",
		name = L[" "],
		order = 61,
	},	
	["createRemoveHeader"] = { 
	   type = "header",
	   name = L["Create or Remove Layouts"],
	   height = 100,
	   order = 55,
	},
   ["newLayoutName"] = {
      type = "text",
      name = L["Name of layout to create or remove:"],
      desc = L["Name of layout to create or remove (hint: type name then press <enter>)."],
      order = 56,
      get = function () return GridConfigurableLayouts.newLayoutName end,
      set = function (v) 
         GridConfigurableLayouts.newLayoutName = v
         waterfall:Refresh("Grid")
      end, 
   },
   ["createNewLayout"] = {
      type = "execute",
      name = L["Create New Layout"],
      desc = L["Create a new layout."],
      order = 57,
      disabled = function () return (not GridConfigurableLayouts.newLayoutName) end,
      func = function () GridConfigurableLayouts:createLayout(GridConfigurableLayouts.newLayoutName) end,
   },
   ["removeLayout"] = {
      type = "execute",
      name = L["Remove Layout"],
      desc = L["Remove a layout."],
      order = 58,
      disabled = function () return (not GridConfigurableLayouts.layoutNames[GridConfigurableLayouts.newLayoutName]) end,
      func = function ()
         GridConfigurableLayouts:removeLayout(GridConfigurableLayouts.newLayoutName)
      end,
   },
   ["petGroupOptions"] = {
      type = "group",
      name = L["Configure Pet Group"],
      desc = L["Settings that apply to pet groups in this layout."],
      order = 21,
      args = {
        	["enableAllPets"] = {
            type = "toggle",
            name = L["Display all raid pets"],
            desc = L["Display raid pets for all pet classes."],
            order = 61,
            get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["allPets"] end,
            set = function ()
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["allPets"] =
                  not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["allPets"]
               GridConfigurableLayouts:checkUpdateCurrent()
               waterfall:Refresh("Grid")
            end, 
         },
         ["enableHunterPets"] = {
            type = "toggle",
            name = L["Display hunter pets"],
            desc = L["Display raid pets for hunters."],
            disabled = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["allPets"] end, 
            order = 61,
            get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["HUNTER"] end,
            set = function ()
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["HUNTER"] =
                  not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["HUNTER"]
               GridConfigurableLayouts:checkUpdateCurrent()
            end, 
         },
         ["enableWarlockPets"] = {
            type = "toggle",
            name = L["Display warlock pets"],
            desc = L["Display raid pets for warlocks."],
            disabled = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["allPets"] end,
            order = 61,
            get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["WARLOCK"] end,
            set = function ()
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["WARLOCK"] =
                  not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["WARLOCK"]
               GridConfigurableLayouts:checkUpdateCurrent()
            end, 
         },
         ["enableDeathknighPets"] = {
            type = "toggle",
            name = L["Display deathknight pets"],
            desc = L["Display raid pets for deathknights."],
            disabled = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["allPets"] end,
            order = 61,
            get = function () return GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["DEATHKNIGHT"] end,
            set = function ()
               GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["DEATHKNIGHT"] =
                  not GridConfigurableLayouts.db.profile.layouts[GridConfigurableLayouts.db.profile.configLayout].petOptions["DEATHKNIGHT"]
               GridConfigurableLayouts:checkUpdateCurrent()
            end, 
         },
      },
   },
   ["globalOptions"] = {
      type = "group",
      name = L["Global Options"],
      desc = L["Set global options."],
      order = 22,
      args = {
         ["createBordersHeader"] = { 
      	   type = "header",
      	   name = L["Custom Borders"],
      	   order = 60,
      	},
      	["createBorders"] = {
            type = "toggle",
            name = L["Replace Grid Border with Group Borders"],
            desc = L["Enable creation of colored borders around groups. Uses GridLayout border texture.\nHint: play with your padding under Layout."],
            order = 61,
            get = function () return GridConfigurableLayouts.db.profile.createBorders end,
            set = function ()
               GridConfigurableLayouts.db.profile.createBorders = not GridConfigurableLayouts.db.profile.createBorders
               GridConfigurableLayouts:setCreateBorders(GridConfigurableLayouts.db.profile.createBorders)
               GridLayout:ReloadLayout()
            end, 
         },
         ["borderFilterSelection"] = {
            type = "text",
            name = L["Filter to apply border too:"],
            desc = L["Choose filter for which to set color."],
            order = 62,
            get = function () return GridConfigurableLayouts.db.profile.filterToColor end,
            set = function (v)
               GridConfigurableLayouts.db.profile.filterToColor = v
               waterfall:Refresh("Grid")
            end,
            validate = GridConfigurableLayouts.groupTypes,
         },
         ["borderColor"] = {
      		type = "color",
      		name = L["Border Color"],
      		desc = L["Adjust border color and alpha."],
      		order = 63,
      		get = function ()
        			local filter = GridConfigurableLayouts.db.profile.filterToColor
               local settings = GridConfigurableLayouts.db.profile
               return settings.filterColors[filter]["r"], settings.filterColors[filter]["g"], settings.filterColors[filter]["b"], settings.filterColors[filter]["a"]
      		end,
      		set = function (r, g, b, a)
      			local filter = GridConfigurableLayouts.db.profile.filterToColor
               local settings = GridConfigurableLayouts.db.profile
      			settings.filterColors[filter]["r"], settings.filterColors[filter]["g"],settings.filterColors[filter]["b"],settings.filterColors[filter]["a"] = r, g, b, a
               GridLayout:UpdateColor()
            end,
            hasAlpha = true
      	},
         ["spacer7"] = {
      		type = "header",
      		name = L[" "],
      		order = 65,
      	},
      	["spacer8"] = {
      		type = "header",
      		name = L[" "],
      		order = 66,
      	},
      	["spacer9"] = {
      		type = "header",
      		name = L[" "],
      		order = 67,
      	},
      	["otherSettingsHeader"] = { 
      	   type = "header",
      	   name = L["Everything Else"],
      	   order = 68,
      	},
         ["checkOOZFrequency"] = {
      	   type = "range",
       		name = L["Zone change check frequency"],
        	   desc = L["Frequency with which to check for raidmember zone changes."],
            order = 69,
            max = 20,
            min = 1,
            step = 1,
            get = function ()	return GridLayout.db.profile.checkOOZFrequency end,
      	   set = function (v)
         		GridLayout.db.profile.checkOOZFrequency = v
         	   GridConfigurableLayouts:checkUpdateCurrent()
      	   end,
         },
      },
   },
}
---}}}


--{{{ Place Menus into the grid menu system under Layout
GridLayout.options.args["advanced"]["order"] = 98
GridLayout.options.args["GridConfigurableLayouts"] = {
	type = "group",
	name = L["Configure Layouts"],
	desc = L["Options for creating custom layouts."],
   order = 99,
	--disabled = InCombatLockdown,
	args = configurableLayoutsOptions
}
--}}}


---{{{ Initialization, reset, etc., . . .
function GridConfigurableLayouts:OnInitialize()
   self.super.OnInitialize(self)
   for layoutName, _ in pairs(GridConfigurableLayouts.db.profile.layouts) do
      GridConfigurableLayouts.layoutNames[layoutName] = layoutName
   end
end

function GridConfigurableLayouts:OnEnable()
   GridConfigurableLayouts:loadNewConfig(GridConfigurableLayouts.db.profile.configLayout)
   GridConfigurableLayouts:addLayouts()
   LoadLayoutHook:Hook(GridLayout, "ReloadLayout")
   GridLayout:ReloadLayout()
end

function GridConfigurableLayouts:Reset()
   self.super.Reset(self)
   GridLayout:ReloadLayout()
end
---}}}


---{{{ Registrations and Callback functions
function GridConfigurableLayouts:doRegister()
   GridConfigurableLayouts:RegisterEvent("Grid_EnteringCombat", GridConfigurableLayouts.setCombatLockdown)
   GridConfigurableLayouts:RegisterEvent("Grid_LeavingCombat", GridConfigurableLayouts.removeCombatLockdown)
   GridConfigurableLayouts:RegisterEvent("Grid_RosterUpdated", GridConfigurableLayouts.updateLayout)
   talentLib.RegisterCallback(GridConfigurableLayouts, 'LibGroupTalents_RoleChange')
   if not combatLockdown and GridConfigurableLayouts:checkUsingOOZ() then
      GridConfigurableLayouts:ScheduleRepeatingEvent("GridConfigurableLayouts_OOZ_Check", GridConfigurableLayouts.OOZCheck, GridLayout.db.profile.checkOOZFrequency )
   end   
   isRegistered = true
end


function GridConfigurableLayouts:doUnregister()
   GridConfigurableLayouts:UnregisterEvent("Grid_EnteringCombat")
   GridConfigurableLayouts:UnregisterEvent("Grid_LeavingCombat")
   GridConfigurableLayouts:UnregisterEvent("Grid_RosterUpdated")
   talentLib.UnregisterCallback(GridConfigurableLayouts, "LibGroupTalents_RoleChange")
   GridConfigurableLayouts:CancelScheduledEvent("GridConfigurableLayouts_OOZ_Check")
   isRegistered = false
end

function GridConfigurableLayouts:setCombatLockdown()
   combatLockdown = true
   GridConfigurableLayouts:CancelScheduledEvent("GridConfigurableLayouts_OOZ_Check")
end


function GridConfigurableLayouts:removeCombatLockdown()
   combatLockdown = false
   if GridConfigurableLayouts:isActive() and GridConfigurableLayouts:checkUsingOOZ() then
      GridConfigurableLayouts:ScheduleRepeatingEvent("GridConfigurableLayouts_OOZ_Check", GridConfigurableLayouts.OOZCheck, GridLayout.db.profile.checkOOZFrequency)
   end
   if reloadLayoutQueued then
      GridLayout:ReloadLayout()
   end
end    
    

function LoadLayoutHook:ReloadLayout(self, layoutName)
   if combatLockdown then
      reloadLayoutQueued = true
   else
      if GridConfigurableLayouts:isActive() then
         GridConfigurableLayouts:redoGroups()
      end
	   GridLayout:LoadLayout(self.db.profile.layouts[GridRoster:GetPartyState()])
	end
end


function LoadLayoutHook:UpdateColor()
   -- This hook is used to place group borders whenever GridLayout refreshes it's visual elements
   if GridConfigurableLayouts:isActive() and GridConfigurableLayouts.db.profile.createBorders then
      GridConfigurableLayouts:makeBorders()
	end
end   


function LoadLayoutHook:PlaceGroup(context, layoutGroup, groupNumber)
   -- This hook is used when we need to adjust group placement to make room for group borders
   if GridConfigurableLayouts:isActive() and GridConfigurableLayouts.db.profile.createBorders then
      GridConfigurableLayouts:adjustGroup(layoutGroup, groupNumber)
	end
end   


function GridConfigurableLayouts:updateLayout()
      GridLayout:ReloadLayout()
end
                        

function GridConfigurableLayouts:OOZCheck()
   -- Periodically checks if any raid member has changed zones
   if not combatLockdown then
      if GridConfigurableLayouts:isActive() then
         for guid, unit in GridRoster:IterateRoster() do
            if (string.sub(unit, 1, 7) ~= "raidpet") then
               local zone = GetRealZoneText()
               if (string.sub(unit, 1, 4) == "raid") then
                  local raidIndex = string.sub(unit, 5)
                  local _, _, _, _, _, _, temp_zone = GetRaidRosterInfo(raidIndex)
                  zone = temp_zone or zone
      	      end
               if rosterZoneList[guid] ~= zone and (zone ~= "Offline" or currentLayout.offlineIsOOZ) then
                  GridLayout:ReloadLayout()
                  return
               end
            end
         end
      end
   end
end   


function GridConfigurableLayouts:LibGroupTalents_RoleChange(event, guid, unit)
   GridLayout:ReloadLayout()
end
---}}}


---{{{ Core functions
function GridConfigurableLayouts:createLayout(name)
   -- Occurs when "Create New Layout" button is pressed, creates a copy of GCL_Default with a new name
   if name and name ~= "" then
      -- Note that if user "Creates" a name that exists in GridLayoutLayouts, it will be taken over by this module!!
      if GridLayout.layoutSettings[name] == nil then
         GridLayout:AddLayout(name, emptyLayout)
      else
         GridLayout.layoutSettings[name] = emptyLayout
      end
      if GridConfigurableLayouts.db.profile.layouts[name] == nil then
         GridConfigurableLayouts.db.profile.layouts[name] = {
            allowRepeats = false,
            numGroups = 6,
            groupType = {"tank", "melee", "healer", "ranged", "pets", "ooz", "none", "none", "none", "none", "none" },
            nameList = {"", "", "", "", "", "", "", "", "", "", "" },
            petOptions = {
               ["allPets"] = false,
               ["HUNTER"] = true,
               ["WARLOCK"] = true,
               ["DEATHKNIGHT"] = true,
            },
         }
         GridConfigurableLayouts.layoutNames[name] = name
      end
      GridConfigurableLayouts:loadNewConfig(name)
      GridConfigurableLayouts.newLayoutName = ""      
      GridConfigurableLayouts:checkUpdateCurrent()
   end
end

function GridConfigurableLayouts:removeLayout(name)
   -- Occurs when "Remove Layout" button is pressed
   if name ~= "GCL Default" and GridConfigurableLayouts.layoutNames[name] then
      GridLayout.layoutSettings[name] = nil
      GridConfigurableLayouts.db.profile.layouts[name] = nil
      GridConfigurableLayouts.layoutNames[name] = nil
      if GridConfigurableLayouts.db.profile.configLayout == name then
         name = "GCL Default"
         GridConfigurableLayouts:loadNewConfig(name)
      end
      GridConfigurableLayouts.newLayoutName = ""
      GridConfigurableLayouts:checkUpdateCurrent()
   end
end   
   
   
function GridConfigurableLayouts:addLayouts()
   -- Adds placeholders for all our layouts, so that they are visible to GridLayout
   for name, _ in pairs(GridConfigurableLayouts.db.profile.layouts) do
      if GridLayout.layoutSettings[name] == nil then
         GridLayout:AddLayout(name, emptyLayout)
      end
   end
end


function GridConfigurableLayouts:redoGroups()
   -- The core function of this module: processes the roster and builds the current layout
   -- which it feeds to GridLayout:LoadLayout
   groupList = {"", "", "", "", "", "", "", "", "", "", "" }
   local petGroupFilter = ""
   local oozList = ""
   local usingooz = GridConfigurableLayouts:checkUsingOOZ()
   local revisedNameList = {}
   local nameListed = {}
   
   -- make a single list of all names in nameLists
   -- and make a clean copy of namelists with no bogus names
   -- (because names that are not in raid can block legit names from appearing)
   for i = 1, currentLayout.numGroups do
      nl = currentLayout.nameList[i] .. ","
      tempnl = ""
      for name in gmatch(nl, "%s*(%w+)%s*,")   do
         if GridRoster:GetGUIDByFullName(name) then
            nameListed[name] = true
            tempnl = tempnl .. "," .. name
         end
      end
      revisedNameList[i] = tempnl
   end
   
   -- iterate thru the roster and place units in groups
   for guid, unit in GridRoster:IterateRoster() do
      if (string.sub(unit, 1, 7) ~= "raidpet") and
            (string.sub(unit, 1, 8) ~= "partypet") and
            (unit ~= "pet") then
         local name, raidgroup, isMT, zone = GridConfigurableLayouts:getUnitInfo(unit, guid)
         local role = GridConfigurableLayouts:getRole(guid, unit)
         unitPlaced = nameListed[name]
         -- first check if this unit belongs in an ooz group
         if usingooz then 
            if zone ~= "Offline" then
               rosterZoneList[guid] = zone
            elseif not currentLayout.offlineIsOOZ then
               zone =  rosterZoneList[guid]
            end
            if zone ~= GetRealZoneText() then
               oozList = oozList .. "," .. name
               unitPlaced = true
            end
         end
         -- then fill the groups in order
         for i = 1, currentLayout.numGroups, 1 do
            if currentLayout.allowRepeats or not unitPlaced then
               if (role == currentLayout.groupType[i]) or
                     (currentLayout.groupType[i] == "blizztank" and isMT) or
                     (currentLayout.groupType[i] == raidgroup) or
                     (currentLayout.groupType[i] == "guild" and UnitIsInMyGuild(unit))
                     then
                  groupList[i] = groupList[i] .. "," .. name
                  unitPlaced = true
               end
            end
         end
      end
   end
   -- if there is an ooz group, fill it
   if usingooz then
      for i = 1, currentLayout.numGroups, 1 do
         if currentLayout.groupType[i] == "ooz" then
            groupList[i] = groupList[i] .. "," .. oozList
         end
      end
   end

   for i = 1, currentLayout.numGroups do
      if revisedNameList[i] ~= "" then
         groupList[i] = groupList[i] .. "," .. revisedNameList[i]
      end
   end
   
   -- create groupFilter for pet group
   if not currentLayout.petOptions["allPets"] then
      for class, enabled in pairs(currentLayout.petOptions) do
         if class ~= "allPets" and enabled == true then
            petGroupFilter = petGroupFilter .. "," .. class
         end
      end
   end
   
   -- time to pass the updated layout to GridLayout
   local layoutName = GridLayout.db.profile.layouts[GridRoster:GetPartyState()]
   local realIndex = 1
   for i = 1, currentLayout.numGroups, 1 do
      GridLayout.layoutSettings[layoutName][i] = nil
      if currentLayout.groupType[i] ~= "none" then
         GridLayout.layoutSettings[layoutName][realIndex] = {}
         if currentLayout.groupType[i] == "pets" then
            GridLayout.layoutSettings[layoutName][realIndex]["isPetGroup"] = true
            GridLayout.layoutSettings[layoutName][realIndex]["groupFilter"] = "1,2,3,4,5,6,7,8"
            GridLayout.layoutSettings[layoutName][realIndex]["nameList"] = ""
            if not currentLayout.petOptions["allPets"] then
               GridLayout.layoutSettings[layoutName][realIndex]["groupFilter"] = petGroupFilter
            end
         else
            GridLayout.layoutSettings[layoutName][realIndex]["isPetGroup"] = false
            GridLayout.layoutSettings[layoutName][realIndex]["groupFilter"] = nil
            GridLayout.layoutSettings[layoutName][realIndex]["nameList"] = groupList[i]
         end
      realIndex = realIndex + 1
      end
   end
end

function GridConfigurableLayouts:makeBorders()
   -- Implements group borders around existing groups by adding a (transparent) bordered frame
   local groupCount = 1
   local groupColors, petGroupColors, totalGroups = GridConfigurableLayouts:getGroupColors()
   local borderTexture = media:Fetch(media.MediaType.BORDER, GridLayout.db.profile.borderTexture)
   
   for _, f in pairs(outerFrames) do
      f:Hide()
   end
   
   for _, f in pairs(outerPetFrames) do
      f:Hide()
   end
   
   for i, layoutGroup in ipairs(GridLayout.layoutGroups) do
		if layoutGroup:IsFrameVisible() and findVisibleUnitFrame(layoutGroup.frame) then
         local f = outerFrames[i] or CreateFrame("Frame", "GCLOuterFrame" .. i, GridLayout.frame)
         local x, y = layoutGroup.frame:GetLeft() - GridLayout.frame:GetLeft(), layoutGroup.frame:GetTop() - GridLayout.frame:GetTop()
         local width, height = layoutGroup:GetFrameWidth(), layoutGroup:GetFrameHeight()
         x = x - 5
         y = y + 5
         width = width + 10
         height = height + 10
         f:SetWidth( width)
         f:SetHeight( height)
         f:SetPoint("TOPLEFT", x, y)
         f:SetBackdrop({
				bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 16,
				edgeFile = borderTexture, edgeSize = 16,
				insets = {left = 4, right = 4, top = 4, bottom = 4},
			})
         local color = groupColors[i]
         f:SetBackdropBorderColor(color.r, color.g, color.b, color.a)
			f:SetBackdropColor(0, 0, 0, 0)
			f:Show()
			outerFrames[i] = f
		end
	end
   for i, layoutGroup in ipairs(GridLayout.layoutPetGroups) do
		if layoutGroup:IsFrameVisible() and findVisibleUnitFrame(layoutGroup.frame) then
         local f = outerPetFrames[i] or CreateFrame("Frame", "GCLOuterPetFrame" .. i, GridLayout.frame)
         local x, y = layoutGroup.frame:GetLeft() - GridLayout.frame:GetLeft(), layoutGroup.frame:GetTop() - GridLayout.frame:GetTop()
         local width, height = layoutGroup:GetFrameWidth(), layoutGroup:GetFrameHeight()
         x = x - 5
         y = y + 5
         width = width + 10
         height = height + 10
         f:SetWidth( width)
         f:SetHeight( height)
         f:SetPoint("TOPLEFT", x, y)
         f:SetBackdrop({
				bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 16,
				edgeFile = borderTexture, edgeSize = 16,
				insets = {left = 4, right = 4, top = 4, bottom = 4},
			})
         local color = petGroupColors[i]
         f:SetBackdropBorderColor(color.r, color.g, color.b, color.a)
			f:SetBackdropColor(0, 0, 0, 0)
			f:Show()
			outerPetFrames[i] = f
		end
	end
end


function GridConfigurableLayouts:adjustGroup(layoutGroup, groupNumber)
   -- For when group borders are active: Adjusts the placement of the group frames
   -- created by GridLayout to make room for our group borders
   local f = layoutGroup.frame
   local GLSettings = GridLayout.db.profile
   local horizontal = GLSettings.horizontal
   local groupAnchor = GLSettings.groupAnchor
   
   local adjustX, adjustY = 0, 0
   
   if horizontal then
      adjustX = 0
      if groupAnchor == "TOPLEFT" or groupAnchor == "TOPRIGHT" then
         adjustY = -10
      else
         adjustY = 10 
      end
   else
      adjustY = 0
      if groupAnchor == "TOPLEFT" or groupAnchor == "BOTTOMLEFT" then
         adjustX = 10
      else
         adjustX = -10
      end
   end
          
   local p, rT, rP, x, y = f:GetPoint(1)
   f:SetPoint(p, rT, rP, x + adjustX, y + adjustY) 
end
---}}}


---{{{ other functions
function GridConfigurableLayouts:isActive()
   -- Checks if the active layout is one of ours, and if so, makes sure we are registered
   -- to react to everything we need to track
   local party_type = GridRoster:GetPartyState()
   currentLayout = GridConfigurableLayouts.db.profile.layouts[GridLayout.db.profile.layouts[party_type]]
   if currentLayout and not isRegistered then
      GridConfigurableLayouts:doRegister()
      if GridConfigurableLayouts.db.profile.createBorders then
         GridConfigurableLayouts:setCreateBorders(GridConfigurableLayouts.db.profile.createBorders)
      end
   elseif not currentLayout and isRegistered then
      GridConfigurableLayouts:doUnregister()
      if GridConfigurableLayouts.db.profile.createBorders then
         GridConfigurableLayouts:setCreateBorders(false)
      end
   end
   return (not (not currentLayout))
end


function GridConfigurableLayouts:checkUpdateCurrent()
   -- When a layout is modified in the options, check if it is the currently active layout
   -- and, if so, reload it with the modified settings
   local party_type = GridRoster:GetPartyState()
   if GridLayout.db.profile.layouts[party_type] == GridConfigurableLayouts.db.profile.configLayout then
      if not combatLockdown and GridConfigurableLayouts:checkUsingOOZ() then
         -- If we are registered we might already have this scheduled.
         -- On the other hand, we may have just added an ooz group or changed the frequency, so . . .
         GridConfigurableLayouts:CancelScheduledEvent("GridConfigurableLayouts_OOZ_Check")
         GridConfigurableLayouts:ScheduleRepeatingEvent("GridConfigurableLayouts_OOZ_Check", GridConfigurableLayouts.OOZCheck, GridLayout.db.profile.checkOOZFrequency)                                                                                                                                                                       
      end
      GridLayout:ReloadLayout()
   end
end            


function GridConfigurableLayouts:getUnitInfo(unit, guid)
   -- The function name says it all
   if (string.sub(unit, 1, 4) == "raid") then
      local raidIndex = string.sub(unit, 5)
      local name, raidgroup, isMT, temp_zone
      local zone = GetRealZoneText()
      name = GridRoster:GetFullNameByGUID(guid)
      _, _, raidgroup, _, _, _, temp_zone, _, _, blizzRole = GetRaidRosterInfo(raidIndex)
      zone = temp_zone or zone -- this deals with a Blizz bug where reloading ui causes GetRaidRosterInfo to return empty zone info for player
      isMT = (blizzRole == "MAINTANK")
      raidgroup = "rg" .. raidgroup
      return name, raidgroup, isMT, zone 
   end
   -- we are solo or in a party
   local name = GridRoster:GetFullNameByGUID(guid)   
   return name, "rg1", false, GetRealZoneText() -- we can't get zone info for party members!?!?
end


function GridConfigurableLayouts:checkUsingOOZ()
   -- Checks if there is an "Out of zone" group in the current layout
   for _, groupType in ipairs(currentLayout.groupType) do
      if groupType == "ooz" then
         return true
      end
   end
   return false
end


function GridConfigurableLayouts:getRole(guid, unit)
   -- Use GroupTalentLib to determine the raid role of a unit
   local _, class = UnitClass(unit)
   local role = (class == "HUNTER" and "ranged") or talentLib:GetGUIDRole(guid)
	if role == "caster" then
      role = "ranged"
	end
	-- if no role was returned use a default until the unit can be inspected
   if role ~= "tank" and role ~= "melee" and role ~= "healer" and role ~= "ranged" then
      if class == "ROGUE" or class == "WARRIOR" or class == "DEATHKNIGHT" then
         role = "melee"
      elseif class == "WARLOCK" or class == "MAGE" or class == "HUNTER" then
         role = "ranged"
      elseif class == "PALADIN" or class == "DRUID" or class == "PRIEST" or class == "SHAMAN" then
         role = "healer"
      end
	end
   return role
end

function GridConfigurableLayouts:getGroupColors()
   -- Used by MakeBorders, this creates tables of group colors that match
   -- the group borders and pet group borders in order of their creation 
   local settings = GridConfigurableLayouts.db.profile
   local i, j, k = 0, 0, 0
   local groupColors = {}
   local petGroupColors = {}
   for _, groupType in pairs(currentLayout.groupType) do
      if groupType == "pets" then
         j = j + 1
         k = k + 1
         petGroupColors[j] =  settings.filterColors["pets"]
         petGroupColors[j]["realIndex"] = k
      elseif groupType == "none" then
         break
      else
         i = i + 1
         k = k + 1
         groupColors[i] = settings.filterColors[groupType]
         groupColors[i]["realIndex"] = k
      end
   end
   return groupColors, petGroupColors, i + j
end


function GridConfigurableLayouts:setCreateBorders(doCreate)
   -- Called when creation of group borders is enabled or disabled
   if doCreate then
      if not LoadLayoutHook:IsHooked(GridLayout, "UpdateColor") then
         LoadLayoutHook:SecureHook(GridLayout, "UpdateColor")
      end
      if not LoadLayoutHook:IsHooked(GridLayout, "PlaceGroup") then
         LoadLayoutHook:SecureHook(GridLayout, "PlaceGroup")
      end
      GridLayout.frame:DisableDrawLayer("BORDER")
      GridLayout.frame:DisableDrawLayer("BACKGROUND")
   else
      if LoadLayoutHook:IsHooked(GridLayout, "PlaceGroup") then
         LoadLayoutHook:Unhook(GridLayout, "PlaceGroup")
      end
      if LoadLayoutHook:IsHooked(GridLayout, "UpdateColor") then
         LoadLayoutHook:Unhook(GridLayout, "UpdateColor")
      end
      for _, f in pairs(outerFrames) do
         f:Hide()
      end
      for _, f in pairs(outerPetFrames) do
         f:Hide()
      end
      local settings = GridLayout.db.profile 
      GridLayout.frame:EnableDrawLayer("BORDER")
      GridLayout.frame:EnableDrawLayer("BACKGROUND")
   end
end

function GridConfigurableLayouts:loadNewConfig(layoutName)
   GridConfigurableLayouts.db.profile.configLayout = layoutName
   GridConfigurableLayouts.allPetsEnabled = GridConfigurableLayouts.db.profile.layouts[layoutName].petOptions["allPets"]
   if not GridConfigurableLayouts.db.profile.layouts[layoutName].numGroups then
      for i = 11, 1, -1 do
         if GridConfigurableLayouts.db.profile.layouts[layoutName].groupType[i] ~= "none" then
            GridConfigurableLayouts.db.profile.layouts[layoutName].numGroups = i
            break
         end
      end
   end
   for i = 11, GridConfigurableLayouts.db.profile.layouts[layoutName].numGroups + 1, -1 do
      GridLayout.options.args["GridConfigurableLayouts"].args["editGroups"].args["groupType" .. i].type = "disabled"
      GridLayout.options.args["GridConfigurableLayouts"].args["editNameLists"].args["nameList" .. i].type = "disabled"      
   end
   for i = 1, GridConfigurableLayouts.db.profile.layouts[layoutName].numGroups, 1 do
      GridLayout.options.args["GridConfigurableLayouts"].args["editGroups"].args["groupType" .. i].type = "text"
      GridLayout.options.args["GridConfigurableLayouts"].args["editNameLists"].args["nameList" .. i].type = "text"
   end
end
---}}}